from .ldap import LDAPProtocol
from .winrm import WinRMProtocol
from .ssh import SSHProtocol


INTERFACES = {
    'ssh': SSHProtocol,
    # 'winexe': WinexeProtocol,
    # 'wmi': WMIProtocol,
    'winrm': WinRMProtocol,
    'ldap': LDAPProtocol,
}


from typing import Optional

import ldap
from ldap.ldapobject import LDAPObject

from ..model import Protocol, Response, Query
from .base import BaseProtocol
from ..exceptions import CredentialsError, ConnectionFailed, ExecutionError


class LDAPProtocol(BaseProtocol):
    name = 'ldap'

    def __init__(self):
        self.base: Optional[str] = None
        self.con: Optional[LDAPObject] = None

    def connect(self, protocol: Protocol) -> LDAPObject:
        server = f'ldap://{protocol.host}'
        ldap_login = f'{protocol.username}@{protocol.domain}'
        self.base = ', '.join(f'dc={i}' for i in protocol.domain.split('.'))
        try:
            self.con = ldap.initialize(server, bytes_mode=False)
            self.con.protocol_version = ldap.VERSION3
            self.con.set_option(ldap.OPT_REFERRALS, 0)
            self.con.simple_bind_s(ldap_login, protocol.password)
            return self.con
        except ldap.INVALID_CREDENTIALS as e:
            raise CredentialsError(str(e))
        except ldap.SERVER_DOWN as e:
            raise ConnectionFailed(e.__str__())
        except Exception as e:
            raise ExecutionError(e.__str__())

    def close(self) -> None:
        try:
            self.con.unbind()
        except Exception as e:
            raise ExecutionError("Неожиданная ошибка при отключении")
        self.con = None

    def execute(self, query: Query) -> Response:
        pass


from typing import List, Union


class SOARException(Exception):
    _return_code: int = 1
    _TEMPLATE = 'Error: {args}.'

    @property
    def return_code(self):
        return self._return_code

    def __init__(self, message: Union[List[str], str, dict]):
        self.message = message
        if isinstance(self.message, List):
            self.message = self._TEMPLATE.format(args=', '.join(self.message))
        elif isinstance(self.message, str):
            self.message = self._TEMPLATE.format(args=self.message)
        else:
            self.message = self._TEMPLATE.format(args=str(self.message))
        super().__init__(self.message)


class DataError(SOARException):
    _return_code: int = 10
    _TEMPLATE = 'Data error: {args}.'


class NoInputError(DataError):
    _return_code: int = 11
    _TEMPLATE = 'Missing input data: {args}.'


class NoSecretsError(DataError):
    _return_code: int = 12
    _TEMPLATE = 'Missing secrets: {args}.'


class BadInputError(DataError):
    _return_code: int = 13
    _TEMPLATE = 'Input data validation error: {args}.'


class BadSecretsError(DataError):
    _return_code: int = 14
    _TEMPLATE = 'Secrets validation error: {args}.'


class ProtocolError(SOARException):
    _return_code: int = 20
    _TEMPLATE = 'Protocol error: {args}.'


class ConnectionFailed(ProtocolError):
    _return_code: int = 21
    _TEMPLATE = 'Connection failed: {args}.'


class CredentialsError(ProtocolError):
    _return_code: int = 22
    _TEMPLATE = 'Invalid credentials: {args}.'


class PermissionsError(ProtocolError):
    _return_code: int = 23
    _TEMPLATE = 'Missing permissions: {args}.'


class ExecutionError(ProtocolError):
    _return_code: int = 24
    _TEMPLATE = 'Command execution error: {args}.'



import re

from ussl.utils import replace_all


def format_output(meta_input_json: dict, input_json: dict) -> dict:
    '''
    Функция предназначена для форматирования строк вывода скрипта/плейбука

        ``meta_input_json``: словарь, содержащий информацию для преобразований;

        ``input_json``: словарь, в котором необходимо совершить преобразования.
    '''
    format_templates: dict = meta_input_json.get('__SOAR.format')
    # Проверка используется ли функция форматирования
    if all([
        format_templates,
        isinstance(format_templates, dict)
    ]):
        # Из каждого шаблона форматирования выделяются форматируемые переменные
        for format_template in format_templates:
            format_variables = re.findall(r'(?<={)(.*?)(?=})', str(format_templates[format_template]))
            if format_variables:
                # Выделяются только те переменные, которые есть в input_json
                format_variables = list(filter(lambda x: x in input_json, format_variables))
                # Выделяются значения переменных
                values = [input_json[i] for i in format_variables]
                # Формируется словарь замены для функции replace_all
                format_dict = dict(zip(['{%s}' % i for i in format_variables], values))
                format_result = replace_all(
                    format_templates[format_template],
                    format_dict)
            else:
                # Если переменных для замены не найдено, то возвращается исходная строка
                format_result = format_templates[format_template]
            # Отформатированная строка добавляется с заданным ключём в input_json
            input_json[format_template] = format_result

    return input_json


def _assign_keys(meta_input_json: dict, input_: dict, target_key: str) -> dict:
    '''
    Функция предназначена для изменения ключей в переданных в неё словарях

        ``meta_input_json``: принимает в себя словарь, содержащий информацию
        для преобразований;

        ``input_``: принимает словарь, в котором необходимо совершить преобразования;

        ``target_key``: принимает в себя строку-указатель на фильтр, по которому
        выделяются и заменяются ключи (secrets/input).
    '''
    replace_dict: dict = meta_input_json.get(f'__SOAR.{target_key}')
    # Проверка используется ли функция замены ключей
    if all([
        replace_dict,
        isinstance(replace_dict, dict),
    ]):
        # Выделение только тех ключей, которые есть в _input
        replacements_keys = [i for i in replace_dict if i in input_]
        # Замена старых ключей на новые
        for old_key in replacements_keys:
            input_[replace_dict[old_key]] = input_.pop(old_key)
    return input_


def assign_input(meta_input_json: dict, input_: dict) -> dict:
    '''
    Функция для замены ключей в input_json

        ``meta_input_json``: принимает в себя словарь, содержащий информацию
        для преобразований;

        ``input_``: принимает словарь, в котором необходимо совершить преобразования;
    '''
    return _assign_keys(meta_input_json, input_, 'input')


def assign_secrets(meta_input_json: dict, input_: dict) -> dict:
    '''
    Функция для замены ключей в secrets

        ``meta_input_json``: принимает в себя словарь, содержащий информацию
        для преобразований;

        ``input_``: принимает словарь, в котором необходимо совершить преобразования;
    '''
    return _assign_keys(meta_input_json, input_, 'secrets')



from typing import Any, Dict, List, Union


def replace_all(text: str, replace_dict: Dict[str, str]) -> str:
    '''
    Функция предназначена для замены множества значений в строке:

        ``text``: исходный текст, в котором необходимо сделать замены;

        ``replace_dict``: словарь, ключём которого является подстрока,
        которую надо заменить, а значением строка, на которую надо
        заменить.

    '''
    for i, j in replace_dict.items():
        text = text.replace(i, j)
    return text


def deep_get(
        obj: Union[dict, list],
        keys: List[Union[int, str]],
        default: Any = None
        ) -> Any:
    '''
    Функция предназначена для безопасного извлечения вложенного значения из объекта:

        ``obj``: объект, из которого необходимо извлечь значение;

        ``keys``: список ключей и индексов, по которым извлекаются вложенные объекты
        и значения;

        ``default``: значение, возвращаемое если не удалось извлечь искомое значений.
    '''
    for key in keys:
        if isinstance(obj, dict):
            obj = obj.get(key, default)
        elif isinstance(obj, list):
            try:
                obj = obj[key]
            except Exception:
                return default
        else:
            return default
    return obj



from ..model import Response, Query, Protocol


class BaseProtocol:

    def connect(self, protocol: Protocol) -> None:
        raise NotImplementedError('Метод connect не реализован')

    def execute(self, query: Query) -> Response:
        raise NotImplementedError('Метод execute не реализован')

    def close(self) -> None:
        raise NotImplementedError('Метод close не реализован')



import io
import re
import socket
import time
from paramiko import AutoAddPolicy, Channel, SSHClient, RSAKey
from paramiko.config import SSH_PORT
from paramiko.ssh_exception import SSHException


from ..model import Query, Response, Protocol
from .base import BaseProtocol
from ..exceptions import ConnectionFailed, ExecutionError


class SSHProtocol(BaseProtocol):
    name = 'ssh'

    def __init__(self) -> None:
        self._chan: Channel
        self._client = SSHClient()
        self._client.set_missing_host_key_policy(AutoAddPolicy())

    def _clean_channel(self) -> None:
        prev_timeout = self._chan.gettimeout()
        self._chan.settimeout(self._clean_timeout)
        # pylint: disable=too-many-nested-blocks
        try:
            while self._chan.get_transport().is_active():
                try:
                    tmp = self._chan.recv(1024)
                    if len(tmp) == 0:
                        break
                except socket.timeout:
                    break
        finally:
            self._chan.settimeout(prev_timeout)

    def connect(
            self,
            protocol: Protocol,
            ) -> None:
        self._encoding = protocol.default('encoding', 'utf-8')
        self._clean_timeout = protocol.default('clean_timeout', 2)
        port = protocol.default('port', SSH_PORT)
        timeout = protocol.default('timeout', 10)
        look_for_keys = protocol.default('look_for_keys', False)
        window_width = protocol.default('window_width', 300)
        auth_timeout = protocol.default('auth_timeout', 60)

        if protocol.password:
            try:
                self._client.connect(
                    protocol.host,
                    port,
                    protocol.username,
                    protocol.password,
                    timeout=timeout,
                    look_for_keys=look_for_keys,
                    auth_timeout=auth_timeout
                )
                self._chan = self._client.invoke_shell(width=window_width)
                self._clean_channel()
            except SSHException as exc:
                raise ConnectionFailed(exc.__str__())

        elif protocol.pem_file:
            try:
                keyfile = io.StringIO(protocol.pem_file)
                self._client.connect(
                    protocol.host,
                    port,
                    protocol.username,
                    pkey=RSAKey.from_private_key(keyfile),
                    timeout=timeout,
                    look_for_keys=look_for_keys,
                    auth_timeout=auth_timeout
                )
                self._chan = self._client.invoke_shell(width=window_width)
                self._clean_channel()
            except SSHException as exc:
                raise ConnectionFailed(exc.__str__())

    def close(self) -> None:
        if self._chan is not None:
            self._chan.close()
        self._client.close()

    def execute(
            self,
            query: Query
            ) -> Response:
        data = b''

        prev_timeout = self._chan.gettimeout()

        if query.sudo is not None:
            try:
                self._chan.send(f'{query.command}\n')
                self._chan.send(f'{query.sudo}\n')
            except SSHException as exc:
                raise ExecutionError(
                    {"command": query.command, "error": exc.__str__()}
                ) from exc
        else:
            try:
                self._chan.send(f'{query.command}\n')
            except SSHException as exc:
                raise ExecutionError(
                    {"command": query.command, "error": exc.__str__()}
                ) from exc

        while True:
            self._chan.settimeout(query.timeout)
            try:
                data = data + self._chan.recv(1024)
                time.sleep(0.5)
            except socket.timeout:
                break
            finally:
                self._chan.settimeout(prev_timeout)

        # На линукс (ubuntu) системах помимо результата в ответ попадает лишняя
        # информация (Имя пользователя, сама команда и т.д.). Этот костыль нужен
        # для того чтобы оставить в выводе только результат работы команды.
        if len(data.splitlines()) > 2:
            result_list = data.splitlines()
            del result_list[0], result_list[-1]
            result = ''.join([i.decode(self._encoding) for i in result_list])
        else:
            result = data

        if query.expects is not None:
            for pattern in query.expects:
                if re.search(pattern, result):
                    return Response(
                        result=result,
                        text=f'Команда {query.command} выполнена успешно',
                        status=True)

            return Response(
                result=result,
                text=f'Команда {query.command} вернула неожиданный ответ',
                status=False)
        else:
            return Response(
                result=result,
                text=f'Команда {query.command} выполнена успешно',
                status=True)



from dataclasses import dataclass
from typing import Any, List, Union, Optional


@dataclass
class Response:
    '''
    Класс, описывающий ответ от конечной системы.

        ``result``: содержит исходный ответ от целевой системы;
        ``text``: содержится форматированный ответ от целевой системы;
        ``status``: содержится статус выполнения переданной команды.
    '''
    result: str
    text: Optional[str] = None
    status: Optional[bool] = None


@dataclass
class Query:
    '''
    Класс, описывающий запросы, передаваемые конечной системе.

        ``command``: содержит командe, которую необходимо выполнить;
        ``timeout``: содержит время, отведенное на выпонение команды;
        ``expects``: содержит регулярные выражения, которые описывают
        ожидаемый ответ от конечной системы;
        ``sudo``: содержит пароль от супер пользователя или enable.
    '''
    command: str
    timeout: int
    expects: Optional[Union[List[str], str]] = None
    sudo: Optional[str] = None


@dataclass
class Protocol:
    '''
    Класс, описывающий протокол для подключения к серверу используя указанный интерфейс.

    Общие для всех интерфейсов поля:
        ``host``: ip-адрес или имя хоста, к которому необходимо подключиться;
        ``username``: имя пользователя, под которым необходимо подключиться;
        ``password``: пароль от указанного пользователя;
        ``interface``: интерфейс, к которому необходимо подключиться (ssh, winrm, и т.д.);
        ``port``: порт, на котором работает интерфейс;
        ``query``: команда или набор команд, которые необходимо выполнить;
        ``encoding``: кодировка запроса;
        ``decoding``: кодировка ответа;
        ``window_width``: ширина окна консоли (влияет на форматирование ответа).

    Поля, специфичные для winrm:
        ``domain``: имя домена к которому необходимо подключиться;
        ``scheme``: схема подключения (http или https);
        ``path``: путь до WS-Management;
        ``transport``: протокол аутентификации.

    Поля, специфичные для ssh:
        ``clean_timeout``: таймаут очищения канала;
        ``look_for_keys``: включить или отключить аутентификацию по ключам;
        ``auth_timeout``: таймаут авторизации;
        ``timeout``: таймаут соединения;
        ``pem_file``: значение закрытого ключа авторизации от указанного пользователя.
    '''
    host: str
    interface: str
    username: str
    domain: str

    password: str = None
    port: Optional[int] = None
    query: Optional[Union[List[Query], Query]] = None
    encoding: Optional[str] = None
    decoding: Optional[str] = None
    window_width: Optional[int] = None

    scheme: Optional[str] = None
    path: Optional[str] = None
    transport: Optional[str] = None

    clean_timeout: Optional[int] = None
    look_for_keys: Optional[bool] = None
    auth_timeout: Optional[int] = None
    timeout: Optional[int] = None
    pem_file: Optional[str] = None

    def default(self, key: str, default: Any = None):
        if self.__getattribute__(key) is not None:
            return self.__getattribute__(key)
        else:
            return default



import sys
import json
import warnings
import pathlib
from marshmallow import Schema, exceptions
from ..exceptions import DataError, SOARException


warnings.filterwarnings("ignore")


class BaseFunction:
    """
    Является базовым классом для всех скриптов, участвующих в обогащении и реагировании.
    При использовании класса необходимо реализовать метод ``function``.
    Автоматически принимаемые значения:
        ``input_json``: Первым аргументом принимает информацию, переданную на вход плейбука;
        ``secrets``: Вторым аргументом приниает секреты.
        ``ensure_ascii``: Указывает, должны ли символы не из набора ASCII быть экранированы. По умолчанию False.
    """
    def __init__(self, inputs_model: Schema = None, secrets_model: Schema = None, ensure_ascii: bool = False) -> None:
        """
        Инициализирует экземпляр класса.

        Args:
            ``ensure_ascii (bool)``: Указывает, должны ли символы не из набора ASCII быть экранированы. По умолчанию False
            ``inputs_model (Schema)``: модель входных данных
            ``secrets_model (Schema)``: модель секретов
        Returns:
            ``None``
        """

        self.inputs_model = inputs_model
        self.secrets_model = secrets_model
        self.ensure_ascii = ensure_ascii
        self._input_json = {}
        self._secrets = {}

        inputs = pathlib.Path(sys.argv[1]).read_text(encoding='utf-8')
        secrets = pathlib.Path(sys.argv[2]).read_text(encoding='utf-8')

        if self.inputs_model is not None:
            self._input_json: dict = self._check_input_data(self.inputs_model, inputs, obj="Input data")
        else:
            self._input_json: dict = json.loads(inputs)

        if self.secrets_model is not None:
            self._secrets: dict = self._check_input_data(self.secrets_model, secrets, obj="Secrets")
        else:
            self._secrets: dict = json.loads(secrets)

        try:
            self.input_json = self.validate_inputs(self._input_json)
            self.secrets = self.validate_secrets(self._secrets)
        except DataError as e:
            self.output(e.__str__(), e.return_code)
        except NotImplementedError:
            self.input_json = self._input_json
            self.secrets = self._secrets

        try:
            result, message = self.function()
        except SOARException as e:
            self.output(e.__str__(), e.return_code)
        else:
            self.output(message, **result)

    def _check_input_data(self, model: Schema, data: str, obj: str) -> dict:
        """
        Проверка входных данных
        """
        try:
            checked_data = model.loads(data)
            return checked_data
        except exceptions.ValidationError as e:
            if isinstance(e.messages, dict):
                self.output(f"{obj} validation error", 1, **{"fields": e.messages})
            else:
                self.output(f"{obj} validation error: {e.__str__()}", 1)

    def validate_inputs(self, input_json: dict) -> dict:
        """
        Метод для дополнительной валидации входных данных.

        При ошибке валидации выбрасывает ValidationError
        """
        raise NotImplementedError

    def validate_secrets(self, secrets: dict) -> dict:
        """
        Метод для дополнительной валидации секретов.

        При ошибке валидации выбрасывает ValidationError
        """
        raise NotImplementedError

    def function(self) -> (dict, str):
        """
        В этом методе необходимо реализовать функцию по обогащению
        или реагированию.

        Методу доступны переменные input_json и secrets.

        Для получения данных используйте переменные input_json и secrets класса BaseFunction.
        Для вывода ошибок необходимо использовать исключения из модуля exceptions.
        Returns:
            (dict, str): Результат обогащения или реагирования и сообщение о результате.
        """
        raise NotImplementedError('Метод function не реализован')

    def output(self,
               message: str,
               return_code: int = 0,
               **kwargs) -> None:
        """
        Выводит результат работы скрипта в формате JSON.

        Args:
            ``message (str)``: Сообщение о результате выполнения скрипта, которое будет выведено.
            ``return_code (int)``: Код возврата, указывающий на успешное выполнение (0) или ошибку (ненулевое значение).
            ``**kwargs``: Дополнительные именованные аргументы. Например, результат сбора данных.

        Returns:
            ``None``
        """
        # Обновляем входной JSON с результатом или сообщением об ошибке
        self._input_json['error' if return_code else 'result'] = message

        # Обновляем входной JSON новыми аргументами
        self._input_json.update(kwargs)

        # Выводим входной JSON в форматированном виде
        print(json.dumps(self._input_json, ensure_ascii=self.ensure_ascii))

        if self._input_json.get("result"):
            del self._input_json["result"]
        # Завершаем выполнение скрипта с кодом 0 в случае успешного выполнения или ненулевым в случае ошибки
        exit(return_code)



from typing import Union

from ..protocol import (
    INTERFACES,
    base
)
from ..model import Protocol, Response, Query
from ..exceptions import ConnectionFailed


class Transport:
    def __init__(self) -> None:
        pass

    def _connect(self, protocol: Protocol):
        if not protocol.interface:
            raise ConnectionFailed('Do not specify the interface on which you want to communicate')
        if protocol.interface not in INTERFACES:
            raise ConnectionFailed('Specified interface not supported')

        self.proto: base.BaseProtocol = INTERFACES[protocol.interface]()
        self.proto.connect(protocol)

    def _execute(self, query: Union[Query, list]) -> Response:
        try:
            if isinstance(query, list):
                for q in range(len(query)):
                    response: Response = self.proto.execute(query[q])
                    if response.status is False:
                        return response
            else:
                response: Response = self.proto.execute(query)

            return response
        finally:
            self.proto.close()

    def connect(self, protocol: Protocol):
        self._connect(protocol)

    def execute(self, query: Union[dict, list]) -> Response:
        return self._execute(query)

    def connect_and_execute(self, protocol: Protocol) -> Response:
        query = protocol.query
        self._connect(protocol)
        return self._execute(query)
